from .api import (
    ConsumeOutputEvent,
    CreatePipeline,
    CreateSecret,
    CreateSpace,
    FunctionEnvironments,
    GetDetailedSpacePipeline,
    ListAccessTokens,
    ListPipelines,
    ListSpaceScopes,
    Pipeline,
    PipelineFunctionOutput,
    PipelineState,
    SinkConnector,
    SourceConnector,
    Space,
)

__all__ = [
    "CreateSpace",
    "Space",
    "ConsumeOutputEvent",
    "CreateSecret",
    "Pipeline",
    "GetDetailedSpacePipeline",
    "ListAccessTokens",
    "SourceConnector",
    "SinkConnector",
    "CreatePipeline",
    "PipelineState",
    "PipelineFunctionOutput",
    "ListSpaceScopes",
    "ListPipelines",
    "FunctionEnvironments",
]



from pydantic import BaseModel


class Error(BaseModel):
    """Bad request error response

    Attributes:
        detail: A message describing the error
    """

    detail: str

    def __str__(self) -> str:
        return self.model_dump_json()



import random
import time

from .api_client import APIClient
from .models import errors, responses


class PipelineDataClient(APIClient):
    """Base Client object to publish and consume events from the given pipeline.

    Attributes:
        glassflow_config (GlassFlowConfig): GlassFlowConfig object to interact
            with GlassFlow API
        pipeline_id (str): The pipeline id to interact with
        pipeline_access_token (str): The access token to access the pipeline
    """

    def __init__(self, pipeline_id: str, pipeline_access_token: str):
        super().__init__()
        self.pipeline_id = pipeline_id
        self.pipeline_access_token = pipeline_access_token
        self.headers = {"X-PIPELINE-ACCESS-TOKEN": self.pipeline_access_token}
        self.query_params = {}

    def validate_credentials(self) -> None:
        """
        Check if the pipeline credentials are valid and raise an error if not
        """

        endpoint = f"/pipelines/{self.pipeline_id}/status/access_token"
        return self._request(method="GET", endpoint=endpoint)

    def _request(
        self,
        method,
        endpoint,
        request_headers=None,
        json=None,
        request_query_params=None,
        files=None,
        data=None,
    ):
        headers = {**self.headers, **(request_headers or {})}
        query_params = {**self.query_params, **(request_query_params or {})}
        try:
            return super()._request(
                method=method,
                endpoint=endpoint,
                request_headers=headers,
                json=json,
                request_query_params=query_params,
                files=files,
                data=data,
            )
        except errors.UnknownError as e:
            if e.status_code == 401:
                raise errors.PipelineAccessTokenInvalidError(e.raw_response) from e
            if e.status_code == 404:
                raise errors.PipelineNotFoundError(
                    self.pipeline_id, e.raw_response
                ) from e
            if e.status_code == 429:
                return errors.PipelineTooManyRequestsError(e.raw_response)
            raise e


class PipelineDataSource(PipelineDataClient):
    def publish(self, request_body: dict) -> responses.PublishEventResponse:
        """Push a new message into the pipeline

        Args:
            request_body: The message to be published into the pipeline

        Returns:
            Response object containing the status code and the raw response

        Raises:
            errors.ClientError: If an error occurred while publishing the event
        """
        endpoint = f"/pipelines/{self.pipeline_id}/topics/input/events"
        http_res = self._request(method="POST", endpoint=endpoint, json=request_body)
        return responses.PublishEventResponse(
            status_code=http_res.status_code,
        )


class PipelineDataSink(PipelineDataClient):
    def __init__(self, pipeline_id: str, pipeline_access_token: str):
        super().__init__(pipeline_id, pipeline_access_token)

        # retry delay for consuming messages (in seconds)
        self._consume_retry_delay_minimum = 1
        self._consume_retry_delay_current = 1
        self._consume_retry_delay_max = 60

    def consume(self) -> responses.ConsumeEventResponse:
        """Consume the last message from the pipeline

        Returns:
            Response object containing the status code and the raw response

        Raises:
            errors.ClientError: If an error occurred while consuming the event

        """

        endpoint = f"/pipelines/{self.pipeline_id}/topics/output/events/consume"
        self._respect_retry_delay()
        http_res = self._request(method="POST", endpoint=endpoint)
        self._update_retry_delay(http_res.status_code)

        body = None
        if http_res.status_code == 200:
            body = http_res.json()
            self._consume_retry_delay_current = self._consume_retry_delay_minimum

        return responses.ConsumeEventResponse(
            status_code=http_res.status_code, body=body
        )

    def consume_failed(self) -> responses.ConsumeFailedResponse:
        """Consume the failed message from the pipeline

        Returns:
            Response object containing the status code and the raw response

        Raises:
            errors.ClientError: If an error occurred while consuming the event

        """

        self._respect_retry_delay()
        endpoint = f"/pipelines/{self.pipeline_id}/topics/failed/events/consume"
        http_res = self._request(method="POST", endpoint=endpoint)

        self._update_retry_delay(http_res.status_code)
        body = None
        if http_res.status_code == 200:
            body = http_res.json()
            self._consume_retry_delay_current = self._consume_retry_delay_minimum

        return responses.ConsumeFailedResponse(
            status_code=http_res.status_code, body=body
        )

    def _update_retry_delay(self, status_code: int):
        if status_code == 200:
            self._consume_retry_delay_current = self._consume_retry_delay_minimum
        elif status_code == 204 or status_code == 429:
            self._consume_retry_delay_current *= 2
            self._consume_retry_delay_current = min(
                self._consume_retry_delay_current, self._consume_retry_delay_max
            )
            self._consume_retry_delay_current += random.uniform(0, 0.1)

    def _respect_retry_delay(self):
        if self._consume_retry_delay_current > self._consume_retry_delay_minimum:
            # sleep before making the request
            time.sleep(self._consume_retry_delay_current)



from .client import GlassFlowClient as GlassFlowClient
from .config import GlassFlowConfig as GlassFlowConfig
from .models import api as internal  # noqa: F401
from .models import errors as errors
from .models import responses as responses
from .pipeline import Pipeline as Pipeline
from .pipeline_data import PipelineDataSink as PipelineDataSink
from .pipeline_data import PipelineDataSource as PipelineDataSource
from .secret import Secret as Secret
from .space import Space as Space



from datetime import datetime

from pydantic import BaseModel


class Space(BaseModel):
    """
    Space response object.

    Attributes:
        name (str): Space name.
        id (int): Space id.
        created_at (datetime): Space creation date.
        permission (str): Space permission.
    """

    name: str
    id: str
    created_at: datetime
    permission: str


class ListSpacesResponse(BaseModel):
    """
    Response from list spaces endpoint.

    Attributes:
        total_amount (int): Total amount of spaces.
        spaces (list[Space]): List of spaces.
    """

    total_amount: int
    spaces: list[Space]



from __future__ import annotations

from enum import Enum
from typing import Any

from pydantic import AwareDatetime, BaseModel, ConfigDict


class Payload(BaseModel):
    """
    Logs payload response object.

    Attributes:
        message (str): log message
    """

    model_config = ConfigDict(
        extra="allow",
    )
    message: str


class FunctionLogEntry(BaseModel):
    """
    Logs entry response object.

    Attributes:
        level (int): Log level.
        severity_code (int): Log severity code.
        timestamp (AwareDatetime): Log timestamp.
        payload (Payload): Log payload.
    """

    level: str
    severity_code: int
    timestamp: AwareDatetime
    payload: Payload


class FunctionLogsResponse(BaseModel):
    """
    Response for a function's logs endpoint.

    Attributes:
        logs (list[FunctionLogEntry]): list of logs
        next (str): ID used to retrieve next page of logs
    """

    logs: list[FunctionLogEntry]
    next: str


class EventContext(BaseModel):
    """
    Event context response object.

    Attributes:
        request_id (str): Request ID.
        external_id (str): External ID.
        receive_time (AwareDatetime): Receive time.
    """

    request_id: str
    external_id: str | None = None
    receive_time: AwareDatetime


class ConsumeOutputEvent(BaseModel):
    """
    Consume output event

    Attributes:
        payload (Any): Payload
        event_context (EventContext): Event context
        status (str): Status
        response (Any): request response
        error_details (str): Error details
        stack_trace (str): Error Stack trace
    """

    payload: Any
    event_context: EventContext
    status: str
    response: Any | None = None
    error_details: str | None = None
    stack_trace: str | None = None


class TestFunctionResponse(ConsumeOutputEvent):
    """Response for Test function endpoint."""

    pass


class BasePipeline(BaseModel):
    """
    Base pipeline response object.

    Attributes:
        name (str): Pipeline name.
        space_id (str): Space ID.
        metadata (dict[str, Any]): Pipeline metadata.
    """

    name: str
    space_id: str
    metadata: dict[str, Any]


class PipelineState(str, Enum):
    """
    Pipeline state
    """

    running = "running"
    paused = "paused"


class Pipeline(BasePipeline):
    """
    Pipeline response object.

    Attributes:
        id (str): Pipeline id
        created_at (AwareDatetime): Pipeline creation time
        state (PipelineState): Pipeline state
    """

    id: str
    created_at: AwareDatetime
    state: PipelineState


class SpacePipeline(Pipeline):
    """
    Pipeline with space response object.

    Attributes:
        space_name (str): Space name
    """

    space_name: str


class ListPipelinesResponse(BaseModel):
    """
    Response for list pipelines endpoint

    Attributes:
        total_amount (int): Total amount of pipelines.
        pipelines (list[SpacePipeline]): List of pipelines.
    """

    total_amount: int
    pipelines: list[SpacePipeline]


class ConsumeEventResponse(BaseModel):
    """
    Response from consume event

    Attributes:
        status_code (int): HTTP status code
        body (ConsumeOutputEvent): Body of the response
    """

    body: ConsumeOutputEvent | None = None
    status_code: int | None = None

    def event(self) -> Any:
        """Return event response."""
        if self.body:
            return self.body.response
        return None


class PublishEventResponseBody(BaseModel):
    """Message pushed to the pipeline."""

    pass


class PublishEventResponse(BaseModel):
    """
    Response from publishing event

    Attributes:
        status_code (int): HTTP status code
    """

    status_code: int | None = None


class ConsumeFailedResponse(BaseModel):
    """
    Response from consuming failed event

    Attributes:
        status_code (int): HTTP status code
        body (ConsumeOutputEvent | None): ConsumeOutputEvent
    """

    body: ConsumeOutputEvent | None = None
    status_code: int | None = None

    def event(self) -> Any:
        """Return failed event response."""
        if self.body:
            return self.body.response
        return None


class AccessToken(BaseModel):
    """
    Access Token response object.

    Attributes:
        id (str): The access token id.
        name (str): The access token name.
        token (str): The access token string.
        created_at (AwareDatetime): The access token creation date.
    """

    id: str
    name: str
    token: str
    created_at: AwareDatetime


class ListAccessTokensResponse(BaseModel):
    """
    Response for listing access tokens endpoint.

    Attributes:
        total_amount (int): Total amount of access tokens.
        access_tokens (list[AccessToken]): List of access tokens.
    """

    access_tokens: list[AccessToken]
    total_amount: int



from importlib.metadata import version

from pydantic import BaseModel


class GlassFlowConfig(BaseModel):
    """Configuration object for GlassFlowClient

    Attributes:
        server_url: The base URL of the GlassFlow API
        sdk_version: The version of the GlassFlow Python SDK
        user_agent: The user agent to be used in the requests

    """

    server_url: str = "https://api.glassflow.dev/v1"
    sdk_version: str = version("glassflow")
    user_agent: str = f"glassflow-python-sdk/{sdk_version}"
    glassflow_client: str = f"python-sdk/{sdk_version}"



from .clienterror import (
    ClientError,
    UnauthorizedError,
    UnknownContentTypeError,
    UnknownError,
)
from .error import Error
from .pipeline import (
    ConnectorConfigValueError,
    PipelineAccessTokenInvalidError,
    PipelineArtifactStillInProgressError,
    PipelineNotFoundError,
    PipelineTooManyRequestsError,
    PipelineUnauthorizedError,
)
from .secret import (
    SecretInvalidKeyError,
    SecretNotFoundError,
    SecretUnauthorizedError,
)
from .space import (
    SpaceIsNotEmptyError,
    SpaceNotFoundError,
    SpaceUnauthorizedError,
)

__all__ = [
    "Error",
    "ClientError",
    "UnknownContentTypeError",
    "UnauthorizedError",
    "ConnectorConfigValueError",
    "SecretInvalidKeyError",
    "SecretNotFoundError",
    "SecretUnauthorizedError",
    "SpaceNotFoundError",
    "SpaceIsNotEmptyError",
    "SpaceUnauthorizedError",
    "PipelineArtifactStillInProgressError",
    "PipelineNotFoundError",
    "PipelineAccessTokenInvalidError",
    "PipelineAccessTokenInvalidError",
    "PipelineTooManyRequestsError",
    "PipelineUnauthorizedError",
    "UnknownError",
]



import requests as requests_http


class ClientError(Exception):
    """Represents an error returned by the API.

    Attributes:
        detail: A message describing the error
        status_code: The status code of the response
        body: The response body
        raw_response: The raw response object

    """

    detail: str
    status_code: int
    body: str
    raw_response: requests_http.Response

    def __init__(
        self,
        detail: str,
        status_code: int,
        body: str,
        raw_response: requests_http.Response,
    ):
        """Create a new ClientError object

        Args:
            detail: A message describing the error
            status_code: The status code of the response
            body: The response body
            raw_response: The raw response object
        """
        self.detail = detail
        self.status_code = status_code
        self.body = body
        self.raw_response = raw_response

    def __str__(self) -> str:
        """Return a string representation of the error

        Returns:
            str: The string representation of the error

        """
        body = ""
        if len(self.body) > 0:
            body = f"\n{self.body}"

        return f"{self.detail}: Status {self.status_code}{body}"


class UnauthorizedError(ClientError):
    """Error caused by a user not authorized."""

    def __init__(self, raw_response: requests_http.Response):
        super().__init__(
            detail="Unauthorized request, Personal Access Token used is invalid",
            status_code=401,
            body=raw_response.text,
            raw_response=raw_response,
        )


class UnknownContentTypeError(ClientError):
    """Error caused by an unknown content type response."""

    def __init__(self, raw_response: requests_http.Response):
        content_type = raw_response.headers.get("Content-Type")
        super().__init__(
            detail=f"unknown content-type received: {content_type}",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )


class UnknownError(ClientError):
    """Error caused by an unknown error."""

    def __init__(self, raw_response: requests_http.Response):
        super().__init__(
            detail="Error in getting response from GlassFlow",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )



from __future__ import annotations

import datetime

from .client import APIClient
from .models import api, errors


class Space(APIClient):
    def __init__(
        self,
        personal_access_token: str,
        name: str | None = None,
        id: str | None = None,
        created_at: datetime.datetime | None = None,
        organization_id: str | None = None,
    ):
        """Creates a new GlassFlow space object

        Args:
            personal_access_token: The personal access token to authenticate
                against GlassFlow
            name: Name of the space
            id: ID of the GlassFlow Space you want to create the pipeline in
            created_at: Timestamp when the space was created

        """
        super().__init__()
        self.name = name
        self.id = id
        self.created_at = created_at
        self.organization_id = organization_id
        self.personal_access_token = personal_access_token
        self.headers = {"Personal-Access-Token": self.personal_access_token}
        self.query_params = {"organization_id": self.organization_id}

    def create(self) -> Space:
        """
        Creates a new GlassFlow space

        Returns:
            Space object

        Raises:
            ValueError: If name is not provided in the constructor

        """
        space_api_obj = api.CreateSpace(name=self.name)

        endpoint = "/spaces"
        http_res = self._request(
            method="POST", endpoint=endpoint, json=space_api_obj.model_dump()
        )

        space_created = api.Space(**http_res.json())
        self.id = space_created.id
        self.created_at = space_created.created_at
        self.name = space_created.name
        return self

    def delete(self) -> None:
        """
        Deletes a GlassFlow space

        Raises:
            ValueError: If ID is not provided in the constructor
            errors.SpaceNotFoundError: If ID provided does not match any
                existing space in GlassFlow
            errors.SpaceUnauthorizedError: If the Personal Access Token is not
                provided or is invalid
            errors.SpaceIsNotEmptyError: If the Space is not empty
        """
        if self.id is None:
            raise ValueError("Space id must be provided in the constructor")

        endpoint = f"/spaces/{self.id}"
        self._request(method="DELETE", endpoint=endpoint)

    def _request(
        self,
        method,
        endpoint,
        request_headers=None,
        json=None,
        request_query_params=None,
        files=None,
        data=None,
    ):
        headers = {**self.headers, **(request_headers or {})}
        query_params = {**self.query_params, **(request_query_params or {})}
        try:
            return super()._request(
                method=method,
                endpoint=endpoint,
                request_headers=headers,
                json=json,
                request_query_params=query_params,
                files=files,
                data=data,
            )
        except errors.UnknownError as e:
            if e.status_code == 401:
                raise errors.SpaceUnauthorizedError(self.id, e.raw_response) from e
            if e.status_code == 404:
                raise errors.SpaceNotFoundError(self.id, e.raw_response) from e
            if e.status_code == 409:
                raise errors.SpaceIsNotEmptyError(e.raw_response) from e
            raise e



from .clienterror import ClientError, requests_http


class ConnectorConfigValueError(Exception):
    """Value error for missing connector settings."""

    def __init__(self, connector_type: str):
        super().__init__(
            f"{connector_type}_kind and {connector_type}_config "
            f"or {connector_type}_config_secret_refs must be provided"
        )


class PipelineNotFoundError(ClientError):
    """Error caused by a pipeline ID not found."""

    def __init__(self, pipeline_id: str, raw_response: requests_http.Response):
        super().__init__(
            detail=f"Pipeline ID {pipeline_id} does not exist",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )


class PipelineUnauthorizedError(ClientError):
    """Pipeline operation not authorized, invalid Personal Access Token"""

    def __init__(self, pipeline_id: str, raw_response: requests_http.Response):
        super().__init__(
            detail=f"Unauthorized request on pipeline {pipeline_id}, "
            f"Personal Access Token used is invalid",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )


class PipelineArtifactStillInProgressError(ClientError):
    """Error returned when the pipeline artifact is still being processed."""

    def __init__(self, pipeline_id: str, raw_response: requests_http.Response):
        super().__init__(
            detail=f"Artifact from pipeline {pipeline_id} "
            f"is still in process, try again later.",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )


class PipelineTooManyRequestsError(ClientError):
    """Error caused by too many requests to a pipeline."""

    def __init__(self, raw_response: requests_http.Response):
        super().__init__(
            detail="Too many requests",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )


class PipelineAccessTokenInvalidError(ClientError):
    """Error caused by invalid access token."""

    def __init__(self, raw_response: requests_http.Response):
        super().__init__(
            detail="The Pipeline Access Token used is invalid",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )



from __future__ import annotations

import sys

import requests as requests_http

from .config import GlassFlowConfig
from .models import errors


class APIClient:
    glassflow_config = GlassFlowConfig()

    def __init__(self):
        """API client constructor"""
        super().__init__()
        self.client = requests_http.Session()

    def _get_core_headers(self) -> dict:
        headers = {
            "Accept": "application/json",
            "Gf-Client": self.glassflow_config.glassflow_client,
            "User-Agent": self.glassflow_config.user_agent,
            "Gf-Python-Version": (
                f"{sys.version_info.major}."
                f"{sys.version_info.minor}."
                f"{sys.version_info.micro}"
            ),
        }
        return headers

    def _request(
        self,
        method,
        endpoint,
        request_headers=None,
        json=None,
        request_query_params=None,
        files=None,
        data=None,
    ):
        headers = self._get_core_headers()
        if request_headers:
            headers.update(request_headers)

        url = self.glassflow_config.server_url + endpoint

        try:
            http_res = self.client.request(
                method,
                url=url,
                params=request_query_params,
                headers=headers,
                json=json,
                files=files,
                data=data,
            )
            http_res.raise_for_status()
        except requests_http.HTTPError as http_err:
            raise errors.UnknownError(http_err.response) from http_err
        return http_res



from __future__ import annotations

from .client import APIClient
from .models import api, errors, operations, responses
from .models.responses.pipeline import AccessToken
from .pipeline_data import PipelineDataSink, PipelineDataSource


class Pipeline(APIClient):
    def __init__(
        self,
        personal_access_token: str,
        name: str | None = None,
        space_id: str | None = None,
        id: str | None = None,
        source_kind: str | None = None,
        source_config: dict | None = None,
        sink_kind: str | None = None,
        sink_config: dict | None = None,
        requirements: str | None = None,
        transformation_file: str | None = None,
        env_vars: list[dict[str, str]] | None = None,
        state: str = "running",
        organization_id: str | None = None,
        metadata: dict | None = None,
        created_at: str | None = None,
    ):
        """Creates a new GlassFlow pipeline object

        Args:
            personal_access_token: The personal access token to authenticate
                against GlassFlow
            id: Pipeline ID
            name: Name of the pipeline
            space_id: ID of the GlassFlow Space you want to create the pipeline in
            transformation_file: Path to file with transformation function of
                the pipeline.
            requirements: Requirements.txt of the pipeline
            source_kind: Kind of source for the pipeline. If no source is
                provided, the default source will be SDK
            source_config: Configuration of the pipeline's source
            sink_kind: Kind of sink for the pipeline. If no sink is provided,
                the default sink will be SDK
            sink_config: Configuration of the pipeline's sink
            env_vars: Environment variables to pass to the pipeline
            state: State of the pipeline after creation.
                It can be either "running" or "paused"
            metadata: Metadata of the pipeline
            created_at: Timestamp when the pipeline was created

        Raises:
            FileNotFoundError: If the transformation file is provided and
                does not exist
        """
        super().__init__()
        self.id = id
        self.name = name
        self.space_id = space_id
        self.personal_access_token = personal_access_token
        self.source_kind = source_kind
        self.source_config = source_config
        self.sink_kind = sink_kind
        self.sink_config = sink_config
        self.requirements = requirements
        self.transformation_code = None
        self.transformation_file = transformation_file
        self.env_vars = env_vars
        self.state = state
        self.organization_id = organization_id
        self.metadata = metadata if metadata is not None else {}
        self.created_at = created_at
        self.access_tokens: list[AccessToken] = []

        self.headers = {"Personal-Access-Token": self.personal_access_token}
        self.query_params = {"organization_id": self.organization_id}
        if self.transformation_file is not None:
            self._read_transformation_file()

        self.source_connector = self._fill_connector(
            "source",
            self.source_kind,
            self.source_config,
        )
        self.sink_connector = self._fill_connector(
            "sink", self.sink_kind, self.sink_config
        )

    def fetch(self) -> Pipeline:
        """
        Fetches pipeline information from the GlassFlow API

        Returns:
            Pipeline object

        Raises:
            ValueError: If ID is not provided in the constructor
            errors.PipelineNotFoundError: If ID provided does not match any
                existing pipeline in GlassFlow
            errors.PipelineUnauthorizedError: If the Personal Access Token is not
                provider or is invalid
        """
        if self.id is None:
            raise ValueError(
                "Pipeline id must be provided in order to fetch it's details"
            )

        endpoint = f"/pipelines/{self.id}"
        http_res = self._request(method="GET", endpoint=endpoint)
        fetched_pipeline = api.GetDetailedSpacePipeline(**http_res.json())
        self._fill_pipeline_details(fetched_pipeline)
        # Fetch Pipeline Access Tokens
        self._list_access_tokens()
        # Fetch function source
        self._get_function_artifact()
        return self

    def create(self) -> Pipeline:
        """
        Creates a new GlassFlow pipeline

        Returns:
            Pipeline object

        Raises:
            ValueError: If name is not provided in the constructor
            ValueError: If space_id is not provided in the constructor
            ValueError: If transformation_file is not provided
                in the constructor
        """

        if self.name is None:
            raise ValueError("Name must be provided in order to create the pipeline")
        if self.space_id is None:
            raise ValueError("Argument space_id must be provided in the constructor")
        if self.transformation_file is None:
            raise ValueError(
                "Argument transformation_file must be provided in the constructor"
            )
        else:
            self._read_transformation_file()

        create_pipeline = api.CreatePipeline(
            name=self.name,
            space_id=self.space_id,
            transformation_function=self.transformation_code,
            requirements_txt=self.requirements,
            source_connector=self.source_connector,
            sink_connector=self.sink_connector,
            environments=self.env_vars,
            state=api.PipelineState(self.state),
            metadata=self.metadata,
        )
        endpoint = "/pipelines"
        http_res = self._request(
            method="POST",
            endpoint=endpoint,
            json=create_pipeline.model_dump(exclude_none=True),
        )
        res_json = http_res.json()
        # using custom operations model because api model does not exist
        res = operations.CreatePipeline(
            **res_json,
        )
        self.id = res.id
        self.created_at = res.created_at
        self.space_id = res.space_id
        self.access_tokens.append(
            AccessToken(
                name="default",
                token=res.access_token,
                id="default",
                created_at=res.created_at,
            )
        )
        return self

    def update(
        self,
        name: str | None = None,
        state: str | None = None,
        transformation_file: str | None = None,
        requirements: str | None = None,
        metadata: dict | None = None,
        source_kind: str | None = None,
        source_config: dict | None = None,
        sink_kind: str | None = None,
        sink_config: dict | None = None,
        env_vars: list[dict[str, str]] | None = None,
    ) -> Pipeline:
        """
        Updates a GlassFlow pipeline

        Args:
            name: Name of the pipeline
            state: State of the pipeline after creation.
                It can be either "running" or "paused"
            transformation_file: Path to file with transformation function of
                the pipeline.
            requirements: Requirements.txt of the pipeline
            source_kind: Kind of source for the pipeline. If no source is
                provided, the default source will be SDK
            source_config: Configuration of the pipeline's source
            sink_kind: Kind of sink for the pipeline. If no sink is provided,
                the default sink will be SDK
            sink_config: Configuration of the pipeline's sink
            env_vars: Environment variables to pass to the pipeline
            metadata: Metadata of the pipeline

        Returns:
            Updated pipeline

        """
        self.fetch()
        if transformation_file is not None or requirements is not None:
            if transformation_file is not None:
                with open(transformation_file) as f:
                    file = f.read()
            else:
                file = self.transformation_code

            if requirements is None:
                requirements = self.requirements

            self._upload_function_artifact(file, requirements)
            self.requirements = requirements
            self.transformation_code = file

        if source_kind is not None:
            source_connector = self._fill_connector(
                "source",
                source_kind,
                source_config,
            )
        else:
            source_connector = self.source_connector

        if sink_kind is not None:
            sink_connector = self._fill_connector("sink", sink_kind, sink_config)
        else:
            sink_connector = self.sink_connector

        if env_vars is not None:
            self._update_function(env_vars)

        # using custom model because api model does not exist
        pipeline_req = operations.UpdatePipelineRequest(
            name=name if name is not None else self.name,
            state=state if state is not None else self.state,
            metadata=metadata if metadata is not None else self.metadata,
            source_connector=source_connector,
            sink_connector=sink_connector,
        )

        endpoint = f"/pipelines/{self.id}"
        body = pipeline_req.model_dump_json(exclude_none=True)
        http_res = self._request(method="PATCH", endpoint=endpoint, data=body)
        # Fetch updated pipeline details and validate
        updated_pipeline = api.GetDetailedSpacePipeline(**http_res.json())
        self._fill_pipeline_details(updated_pipeline)
        return self

    def delete(self) -> None:
        """
        Deletes a GlassFlow pipeline

        Raises:
            ValueError: If ID is not provided in the constructor
            error.PipelineNotFoundError: If ID provided does not match any
                existing pipeline in GlassFlow
            errors.PipelineUnauthorizedError: If the Personal Access Token is not
                provided or is invalid
        """
        if self.id is None:
            raise ValueError("Pipeline id must be provided")

        endpoint = f"/pipelines/{self.id}"
        self._request(method="DELETE", endpoint=endpoint)

    def get_logs(
        self,
        page_size: int = 50,
        page_token: str | None = None,
        severity_code: int = 100,
        start_time: str | None = None,
        end_time: str | None = None,
    ) -> responses.FunctionLogsResponse:
        """
        Get the pipeline's logs

        Args:
            page_size: Pagination size
            page_token: Page token filter (use for pagination)
            severity_code: Severity code filter (100, 200, 300, 400, 500)
            start_time: Start time filter
            end_time: End time filter

        Returns:
            Response with the logs
        """

        query_params = {
            "page_size": page_size,
            "page_token": page_token,
            "severity_code": severity_code,
            "start_time": start_time,
            "end_time": end_time,
        }
        endpoint = f"/pipelines/{self.id}/functions/main/logs"
        http_res = self._request(
            method="GET", endpoint=endpoint, request_query_params=query_params
        )
        base_res_json = http_res.json()
        logs = [responses.FunctionLogEntry(**entry) for entry in base_res_json["logs"]]
        return responses.FunctionLogsResponse(
            logs=logs,
            next=base_res_json["next"],
        )

    def get_source(
        self, pipeline_access_token_name: str | None = None
    ) -> PipelineDataSource:
        """
        Get source client to publish data to the pipeline

        Args:
            pipeline_access_token_name (str | None): Name of the pipeline
                access token to use. If not specified, the default token
                will be used

        Returns:
            Source client to publish data to the pipeline

        Raises:
            ValueError: If pipeline id is not provided in the constructor
        """
        return self._get_data_client("source", pipeline_access_token_name)

    def get_sink(
        self, pipeline_access_token_name: str | None = None
    ) -> PipelineDataSink:
        """
        Get sink client to consume data from the pipeline

        Args:
            pipeline_access_token_name (str | None): Name of the pipeline
                access token to use. If not specified, the default token
                will be used

        Returns:
            Sink client to consume data from the pipeline

        Raises:
            ValueError: If pipeline id is not provided in the constructor
        """
        return self._get_data_client("sink", pipeline_access_token_name)

    def test(self, data: dict) -> responses.TestFunctionResponse:
        """
        Test a pipeline's function with a sample input JSON

        Args:
            data: Input JSON

        Returns:
            Test function response
        """
        endpoint = f"/pipelines/{self.id}/functions/main/test"
        request_body = data
        http_res = self._request(method="POST", endpoint=endpoint, json=request_body)
        base_res_json = http_res.json()
        return responses.TestFunctionResponse(
            **base_res_json,
        )

    def _request(
        self,
        method,
        endpoint,
        request_headers=None,
        json=None,
        request_query_params=None,
        files=None,
        data=None,
    ):
        headers = {**self.headers, **(request_headers or {})}
        query_params = {**self.query_params, **(request_query_params or {})}
        try:
            return super()._request(
                method=method,
                endpoint=endpoint,
                request_headers=headers,
                json=json,
                request_query_params=query_params,
                files=files,
                data=data,
            )
        except errors.UnknownError as e:
            if e.status_code == 401:
                raise errors.PipelineUnauthorizedError(self.id, e.raw_response) from e
            if e.status_code == 404:
                raise errors.PipelineNotFoundError(self.id, e.raw_response) from e
            if e.status_code == 425:
                raise errors.PipelineArtifactStillInProgressError(
                    self.id, e.raw_response
                ) from e
            raise e

    @staticmethod
    def _fill_connector(
        connector_type: str, kind: str, config: dict
    ) -> api.SourceConnector | api.SinkConnector:
        """Format connector input"""
        if not kind and not config:
            connector = None
        elif kind and config:
            connector = dict(kind=kind, config=config)
        else:
            raise errors.ConnectorConfigValueError(connector_type)

        if connector_type == "source":
            return api.SourceConnector(root=connector)
        elif connector_type == "sink":
            return api.SinkConnector(root=connector)
        else:
            raise ValueError("connector_type must be 'source' or 'sink'")

    def _list_access_tokens(self) -> Pipeline:
        endpoint = f"/pipelines/{self.id}/access_tokens"
        http_res = self._request(method="GET", endpoint=endpoint)
        tokens = responses.ListAccessTokensResponse(**http_res.json())
        self.access_tokens = tokens.access_tokens
        return self

    def _get_function_artifact(self) -> Pipeline:
        """
        Fetch pipeline function source

        Returns:
            Pipeline with function source details
        """
        endpoint = f"/pipelines/{self.id}/functions/main/artifacts/latest"
        http_res = self._request(method="GET", endpoint=endpoint)
        res_json = http_res.json()
        self.transformation_code = res_json["transformation_function"]
        # you would never know what else was changed

        if "requirements_txt" in res_json:
            self.requirements = res_json["requirements_txt"]
        return self

    def _upload_function_artifact(self, file: str, requirements: str) -> None:
        files = {"file": file}
        data = {
            "requirementsTxt": requirements,
        }
        endpoint = f"/pipelines/{self.id}/functions/main/artifacts"
        self._request(method="POST", endpoint=endpoint, files=files, data=data)

    def _update_function(self, env_vars):
        """
        Patch pipeline function

        Args:
            env_vars: Environment variables to update

        Returns:
            Pipeline with updated function
        """
        endpoint = f"/pipelines/{self.id}/functions/main"
        body = api.PipelineFunctionOutput(environments=env_vars)
        http_res = self._request(
            method="PATCH", endpoint=endpoint, json=body.model_dump()
        )
        self.env_vars = http_res.json()["environments"]
        return self

    def _get_data_client(
        self, client_type: str, pipeline_access_token_name: str | None = None
    ) -> PipelineDataSource | PipelineDataSink:
        if self.id is None:
            raise ValueError("Pipeline id must be provided in the constructor")
        elif len(self.access_tokens) == 0:
            self._list_access_tokens()

        if pipeline_access_token_name is not None:
            for t in self.access_tokens:
                if t.name == pipeline_access_token_name:
                    token = t.token
                    break
            else:
                raise ValueError(
                    f"Token with name {pipeline_access_token_name} was not found"
                )
        else:
            token = self.access_tokens[0].token
        if client_type == "source":
            client = PipelineDataSource(
                pipeline_id=self.id,
                pipeline_access_token=token,
            )
        elif client_type == "sink":
            client = PipelineDataSink(
                pipeline_id=self.id,
                pipeline_access_token=token,
            )
        else:
            raise ValueError("client_type must be either source or sink")
        return client

    def _read_transformation_file(self):
        try:
            with open(self.transformation_file) as f:
                self.transformation_code = f.read()
        except FileNotFoundError:
            raise

    def _fill_pipeline_details(
        self, pipeline_details: api.GetDetailedSpacePipeline
    ) -> Pipeline:
        self.id = pipeline_details.id
        self.name = pipeline_details.name
        self.space_id = pipeline_details.space_id
        self.state = pipeline_details.state
        source_connector = pipeline_details.source_connector
        if source_connector.root:
            self.source_kind = source_connector.root.kind
            self.source_config = source_connector.root.config
        sink_connector = pipeline_details.sink_connector
        if sink_connector.root:
            self.sink_kind = sink_connector.root.kind
            self.sink_config = sink_connector.root.config
        self.created_at = pipeline_details.created_at
        self.env_vars = pipeline_details.environments

        return self



import click

from .commands import get_started


@click.group()
def glassflow():
    """Glassflow CLI - Manage and control Glassflow SDK"""
    pass


@click.command()
@click.argument("command", required=False)
def help(command):
    """Displays help information about Glassflow CLI and its commands."""

    commands = {
        "get-started": "Initialize Glassflow with an access token.\nUsage: "
        "glassflow get-started --token YOUR_TOKEN",
        "help": "Shows help information.\nUsage: glassflow help [command]",
    }

    if command:
        if command in commands:
            click.echo(f"ℹ️  Help for `{command}`:\n{commands[command]}")
        else:
            click.echo(
                f"❌ Unknown command: `{command}`. Run `glassflow help` for a "
                f"list of commands."
            )
    else:
        click.echo("📖 Glassflow CLI Help:")
        for cmd, desc in commands.items():
            click.echo(f"  ➜ {cmd}: {desc.splitlines()[0]}")
        click.echo("\nRun `glassflow help <command>` for more details.")


# Add commands to CLI group
glassflow.add_command(get_started)
glassflow.add_command(help)

if __name__ == "__main__":
    glassflow()



import os

import click
from dotenv import load_dotenv


@click.command()
@click.option(
    "--personal-access-token", "-pat", default=None, help="Personal access token."
)
@click.option(
    "--env-file",
    "-e",
    default=".env",
    help="Path to the .env file (default: .env in current directory).",
)
def get_started(personal_access_token, env_file):
    """Displays a welcome message and setup instructions."""

    # Load token from .env if not provided in CLI
    if personal_access_token is None:
        if os.path.exists(env_file):
            load_dotenv(env_file)  # Load environment variables
            personal_access_token = os.getenv("PERSONAL_ACCESS_TOKEN")
        else:
            click.echo("⚠️ No token provided and .env file not found!", err=True)
            return

    if not personal_access_token:
        click.echo("❌ Error: Personal access token is required.", err=True)
        return

    click.echo("🚀 Welcome to Glassflow! \n")
    click.echo(
        f"🔑 Using Personal Access Token: {personal_access_token[:4]}... "
        f"(hidden for security)"
    )
    click.echo("\n📝 In this getting started guide, we will do the following:")
    click.echo("1. Define a data transformation function in Python.\n")
    click.echo("2. Create a pipeline with the function.\n")
    click.echo("3. Send events to the pipeline.\n")
    click.echo("4. Consume transformed events in real-time from the pipeline\n")
    click.echo("5. Monitor the pipeline and view logs.\n")

    filename = create_transformation_function()
    pipeline = create_space_pipeline(personal_access_token, filename)
    send_consume_events(pipeline)

    click.echo(
        "\n🎉 Congratulations! You have successfully created a pipeline and sent"
        " events to it.\n"
    )
    click.echo(
        "💻 View the logs and monitor the Pipeline in the "
        f"Glassflow Web App at https://app.glassflow.dev/pipelines/{pipeline.id}"
    )


def create_transformation_function(filename="transform_getting_started.py"):
    file_content = """import json
import logging

def handler(data: dict, log: logging.Logger):
    log.info("Echo: " + json.dumps(data))
    data['transformed_by'] = "glassflow"

    return data
"""
    with open(filename, "w") as f:
        f.write(file_content)
    click.echo(f"✅ Transformation function created in {filename}")
    click.echo("The transformation function is:\n")
    click.echo(file_content)
    click.echo("📝 You can modify the transformation function in the file.")
    return filename


def create_space_pipeline(personal_access_token, transform_filename):
    import glassflow

    # create glassflow client to interact with GlassFlow
    client = glassflow.GlassFlowClient(personal_access_token=personal_access_token)
    example_space = client.create_space(name="getting-started")
    pipeline = client.create_pipeline(
        name="getting-started-pipeline",
        transformation_file=transform_filename,
        space_id=example_space.id,
    )
    click.echo(f"✅ Created a pipeline with pipeline_id {pipeline.id}")
    return pipeline


def send_consume_events(pipeline):
    click.echo("🔄 Sending some generated events to pipeline .....")
    data_source = pipeline.get_source()
    for i in range(10):
        event = {"data": f"hello GF {i}"}
        res = data_source.publish(event)
        if res.status_code == 200:
            click.echo(f"Sent event: {event}")

    click.echo("📡 Consuming transformed events from the pipeline")
    data_sink = pipeline.get_sink()
    for _ in range(10):
        resp = data_sink.consume()
        if resp.status_code == 200:
            click.echo(f"Consumed event: {resp.event()} ")



# ruff: noqa
# generated by datamodel-codegen:
#   filename:  https://api.glassflow.dev/v1/openapi.yaml
#   version:   0.27.2

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field, RootModel


class Error(BaseModel):
    detail: str


class CreateOrganization(BaseModel):
    name: str


class Organization(CreateOrganization):
    id: str


class OrganizationScope(Organization):
    role: str


class OrganizationScopes(RootModel[list[OrganizationScope]]):
    root: list[OrganizationScope]


class SecretKey(RootModel[str]):
    root: str


class Type(str, Enum):
    organization = "organization"


class SecretRef(BaseModel):
    type: Type
    key: SecretKey


class CreateSecret(BaseModel):
    key: SecretKey
    value: str


class SignUp(BaseModel):
    access_token: str
    id_token: str


class BasePipeline(BaseModel):
    name: str
    space_id: str
    metadata: dict[str, Any]


class PipelineState(str, Enum):
    running = "running"
    paused = "paused"


class FunctionEnvironment(BaseModel):
    name: str
    value: str


class FunctionEnvironments(RootModel[Optional[list[FunctionEnvironment]]]):
    root: Optional[list[FunctionEnvironment]] = None


class Kind(str, Enum):
    google_pubsub = "google_pubsub"


class Config(BaseModel):
    project_id: str
    subscription_id: str
    credentials_json: str


class Kind1(str, Enum):
    amazon_sqs = "amazon_sqs"


class Config1(BaseModel):
    queue_url: str
    aws_region: str
    aws_access_key: str
    aws_secret_key: str


class Kind2(str, Enum):
    postgres = "postgres"


class Config2(BaseModel):
    db_host: str
    db_port: Optional[str] = "5432"
    db_user: str
    db_pass: str
    db_name: str
    db_sslmode: Optional[str] = None
    replication_slot: str
    publication: Optional[str] = None
    replication_output_plugin_name: Optional[str] = "wal2json"
    replication_output_plugin_args: Optional[list[str]] = None


class Kind3(str, Enum):
    webhook = "webhook"


class Method(str, Enum):
    get = "GET"
    post = "POST"
    put = "PUT"
    patch = "PATCH"
    delete = "DELETE"


class Header(BaseModel):
    name: str
    value: str


class Config3(BaseModel):
    url: str
    method: Method
    headers: list[Header]


class SinkConnectorWebhookConfigHeadersListItem(BaseModel):
    name: str
    value: str


class SinkConnectorWebhookConfigHeadersList(
    RootModel[list[SinkConnectorWebhookConfigHeadersListItem]]
):
    root: list[SinkConnectorWebhookConfigHeadersListItem]


class Kind4(str, Enum):
    clickhouse = "clickhouse"


class Config4(BaseModel):
    addr: str
    database: str
    username: str
    password: str
    table: str


class Kind5(str, Enum):
    amazon_s3 = "amazon_s3"


class Config5(BaseModel):
    s3_bucket: str
    s3_key: str
    aws_region: str
    aws_access_key: str
    aws_secret_key: str


class Kind6(str, Enum):
    snowflake_cdc_json = "snowflake_cdc_json"


class Config6(BaseModel):
    account: str
    warehouse: str
    db_user: str
    db_pass: str
    db_name: str
    db_schema: str
    db_host: Optional[str] = None
    db_port: Optional[str] = "443"
    db_role: Optional[str] = None


class Kind7(str, Enum):
    pinecone_json = "pinecone_json"


class ClientHeader(BaseModel):
    name: str
    value: str


class Config7(BaseModel):
    api_key: str
    api_host: str
    api_source_tag: Optional[str] = None
    index_host: str
    client_headers: Optional[list[ClientHeader]] = None


class Kind8(str, Enum):
    mongodb_json = "mongodb_json"


class Config8(BaseModel):
    connection_uri: str
    db_name: str


class Pipeline(BasePipeline):
    id: str
    created_at: datetime
    state: PipelineState


class SpacePipeline(Pipeline):
    space_name: str


class PipelineFunctionOutput(BaseModel):
    environments: FunctionEnvironments


class SpacePipelines(RootModel[list[SpacePipeline]]):
    root: list[SpacePipeline]


class CreateSpace(BaseModel):
    name: str


class UpdateSpace(BaseModel):
    name: str


class Space(CreateSpace):
    id: str
    created_at: datetime


class SpaceScope(Space):
    permission: str


class SpaceScopes(RootModel[list[SpaceScope]]):
    root: list[SpaceScope]


class Payload(BaseModel):
    model_config = ConfigDict(
        extra="allow",
    )
    message: str


class SeverityCodeInput(int, Enum):
    integer_100 = 100
    integer_200 = 200
    integer_400 = 400
    integer_500 = 500


class SeverityCode(RootModel[int]):
    root: int


class CreateAccessToken(BaseModel):
    name: str


class AccessToken(CreateAccessToken):
    id: str
    token: str
    created_at: datetime


class AccessTokens(RootModel[list[AccessToken]]):
    root: list[AccessToken]


class PaginationResponse(BaseModel):
    total_amount: int


class SourceFile(BaseModel):
    name: str
    content: str


class SourceFiles(RootModel[list[SourceFile]]):
    root: list[SourceFile]


class EventContext(BaseModel):
    request_id: str
    external_id: Optional[str] = None
    receive_time: datetime


class PersonalAccessToken(RootModel[str]):
    root: str


class QueryRangeMatrix(RootModel[Optional[Any]]):
    root: Optional[Any] = None


class ConnectorValueValue(BaseModel):
    value: str


class ConnectorValueSecretRef(BaseModel):
    secret_ref: SecretRef


class ConnectorValueList(RootModel[list[str]]):
    root: list[str]


class Profile(BaseModel):
    id: str
    home_organization: Organization
    name: str
    email: str
    provider: str
    external_settings: Any
    subscriber_id: str


class ListOrganizationScopes(PaginationResponse):
    organizations: OrganizationScopes


class Secret(BaseModel):
    key: SecretKey


class ListPipelines(PaginationResponse):
    pipelines: SpacePipelines


class ListSpaceScopes(PaginationResponse):
    spaces: SpaceScopes


class FunctionLogEntry(BaseModel):
    level: str
    severity_code: SeverityCode
    timestamp: datetime
    payload: Payload


class ListAccessTokens(PaginationResponse):
    access_tokens: AccessTokens


class ConsumeInputEvent(BaseModel):
    req_id: Optional[str] = Field(None, description="DEPRECATED")
    receive_time: Optional[datetime] = Field(None, description="DEPRECATED")
    payload: Any
    event_context: EventContext


class ConsumeOutputEvent(BaseModel):
    req_id: Optional[str] = Field(None, description="DEPRECATED")
    receive_time: Optional[datetime] = Field(None, description="DEPRECATED")
    payload: Any
    event_context: EventContext
    status: str
    response: Optional[Any] = None
    error_details: Optional[str] = None
    stack_trace: Optional[str] = None


class ListPersonalAccessTokens(BaseModel):
    tokens: list[PersonalAccessToken]


class PipelineInputQueueRelativeLatencyMetricsResponse(BaseModel):
    input_queue_total_push_events: QueryRangeMatrix
    input_queue_latency: QueryRangeMatrix


class ConnectorValue(RootModel[Union[ConnectorValueValue, ConnectorValueSecretRef]]):
    root: Union[ConnectorValueValue, ConnectorValueSecretRef]


class Secrets(RootModel[list[Secret]]):
    root: list[Secret]


class Configuration(BaseModel):
    project_id: ConnectorValue
    subscription_id: ConnectorValue
    credentials_json: ConnectorValue


class SourceConnectorGooglePubSub(BaseModel):
    kind: Literal["google_pubsub"]
    config: Optional[Config] = None
    configuration: Optional[Configuration] = None


class Configuration1(BaseModel):
    queue_url: ConnectorValue
    aws_region: ConnectorValue
    aws_access_key: ConnectorValue
    aws_secret_key: ConnectorValue


class SourceConnectorAmazonSQS(BaseModel):
    kind: Literal["amazon_sqs"]
    config: Optional[Config1] = None
    configuration: Optional[Configuration1] = None


class Configuration2(BaseModel):
    db_host: ConnectorValue
    db_port: Optional[ConnectorValue] = None
    db_user: ConnectorValue
    db_pass: ConnectorValue
    db_name: ConnectorValue
    db_sslmode: Optional[ConnectorValue] = None
    replication_slot: ConnectorValue
    publication: Optional[ConnectorValue] = None
    replication_output_plugin_name: Optional[ConnectorValue] = None
    replication_output_plugin_args: Optional[ConnectorValueList] = None


class SourceConnectorPostgres(BaseModel):
    kind: Literal["postgres"]
    config: Optional[Config2] = None
    configuration: Optional[Configuration2] = None


class Configuration3(BaseModel):
    url: ConnectorValue
    method: ConnectorValue
    headers: SinkConnectorWebhookConfigHeadersList


class SinkConnectorWebhook(BaseModel):
    kind: Literal["webhook"]
    config: Optional[Config3] = None
    configuration: Optional[Configuration3] = None


class Configuration4(BaseModel):
    addr: ConnectorValue
    database: ConnectorValue
    username: ConnectorValue
    password: ConnectorValue
    table: ConnectorValue


class SinkConnectorClickhouse(BaseModel):
    kind: Literal["clickhouse"]
    config: Optional[Config4] = None
    configuration: Optional[Configuration4] = None


class Configuration5(BaseModel):
    s3_bucket: ConnectorValue
    s3_key: ConnectorValue
    aws_region: ConnectorValue
    aws_access_key: ConnectorValue
    aws_secret_key: ConnectorValue


class SinkConnectorAmazonS3(BaseModel):
    kind: Literal["amazon_s3"]
    config: Optional[Config5] = None
    configuration: Optional[Configuration5] = None


class Configuration6(BaseModel):
    account: ConnectorValue
    warehouse: ConnectorValue
    db_user: ConnectorValue
    db_pass: ConnectorValue
    db_name: ConnectorValue
    db_schema: ConnectorValue
    db_host: Optional[ConnectorValue] = None
    db_port: Optional[ConnectorValue] = None
    db_role: Optional[ConnectorValue] = None


class SinkConnectorSnowflakeCDCJSON(BaseModel):
    kind: Literal["snowflake_cdc_json"]
    config: Optional[Config6] = None
    configuration: Optional[Configuration6] = None


class ClientHeader1(BaseModel):
    name: str
    value: ConnectorValue


class Configuration7(BaseModel):
    api_key: ConnectorValue
    api_host: ConnectorValue
    api_source_tag: Optional[ConnectorValue] = None
    index_host: ConnectorValue
    client_headers: Optional[list[ClientHeader1]] = None


class SinkConnectorPineconeJSON(BaseModel):
    kind: Literal["pinecone_json"]
    config: Optional[Config7] = None
    configuration: Optional[Configuration7] = None


class Configuration8(BaseModel):
    connection_uri: ConnectorValue
    db_name: ConnectorValue


class SinkConnectorMongoDBJSON(BaseModel):
    kind: Literal["mongodb_json"]
    config: Optional[Config8] = None
    configuration: Optional[Configuration8] = None


class FunctionLogs(RootModel[list[FunctionLogEntry]]):
    root: list[FunctionLogEntry]


class ListOrganizationSecrets(PaginationResponse):
    secrets: Secrets


class SourceConnector(
    RootModel[
        Optional[
            Union[
                SourceConnectorGooglePubSub,
                SourceConnectorAmazonSQS,
                SourceConnectorPostgres,
            ]
        ]
    ]
):
    root: Optional[
        Union[
            SourceConnectorGooglePubSub,
            SourceConnectorAmazonSQS,
            SourceConnectorPostgres,
        ]
    ] = Field(None, discriminator="kind")


class SinkConnector(
    RootModel[
        Optional[
            Union[
                SinkConnectorWebhook,
                SinkConnectorClickhouse,
                SinkConnectorAmazonS3,
                SinkConnectorSnowflakeCDCJSON,
                SinkConnectorPineconeJSON,
                SinkConnectorMongoDBJSON,
            ]
        ]
    ]
):
    root: Optional[
        Union[
            SinkConnectorWebhook,
            SinkConnectorClickhouse,
            SinkConnectorAmazonS3,
            SinkConnectorSnowflakeCDCJSON,
            SinkConnectorPineconeJSON,
            SinkConnectorMongoDBJSON,
        ]
    ] = Field(None, discriminator="kind")


class GetDetailedSpacePipeline(SpacePipeline):
    source_connector: SourceConnector
    sink_connector: SinkConnector
    environments: FunctionEnvironments


class UpdatePipeline(BaseModel):
    name: str
    transformation_function: Optional[str] = None
    transformation_requirements: Optional[list[str]] = None
    requirements_txt: Optional[str] = None
    metadata: Optional[dict[str, Any]] = None
    source_connector: Optional[SourceConnector] = None
    sink_connector: Optional[SinkConnector] = None
    environments: Optional[FunctionEnvironments] = None


class CreatePipeline(BasePipeline):
    transformation_function: Optional[str] = None
    transformation_requirements: Optional[list[str]] = None
    requirements_txt: Optional[str] = None
    source_connector: Optional[SourceConnector] = None
    sink_connector: Optional[SinkConnector] = None
    environments: Optional[FunctionEnvironments] = None
    state: Optional[PipelineState] = None



import setuptools

try:
    with open("README.md") as fh:
        long_description = fh.read()
except FileNotFoundError:
    long_description = ""

setuptools.setup(
    name="glassflow",
    version="2.1.0",
    author="glassflow",
    description="GlassFlow Python Client SDK",
    url="https://www.glassflow.dev/docs",
    project_urls={
        "Source": "https://github.com/glassflow/glassflow-python-sdk",
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(where="src"),
    install_requires=[
        "urllib3==1.26.15",
        "certifi>=2023.7.22",
        "charset-normalizer>=3.2.0",
        "pydantic>=2.10.6",
        "idna>=3.4",
        "jsonpath-python>=1.0.6 ",
        "marshmallow>=3.19.0",
        "mypy-extensions>=1.0.0",
        "packaging>=23.1",
        "python-dateutil>=2.8.2",
        "requests>=2.31.0",
        "six>=1.16.0",
        "typing-inspect>=0.9.0",
        "typing_extensions>=4.7.1",
        "python-dotenv==1.0.1",
        "eval_type_backport>=0.2.0",
    ],
    extras_require={
        "dev": [
            "pylint>=2.16.2",
            "pytest>=8.3.2",
            "pytest-cov>=5.0.0",
            "datamodel-code-generator[http]>=0.27.0",
            "requests-mock>=1.12.1",
            "isort>=5.13.2",
            "ruff>=0.9.0",
        ]
    },
    package_dir={"": "src"},
    python_requires=">=3.8",
    package_data={"glassflow": ["py.typed"]},
    entry_points={
        "console_scripts": [
            "glassflow = cli.cli:glassflow",
        ],
    },
)



from __future__ import annotations

from pydantic import AwareDatetime, BaseModel

from glassflow.models.api import (
    PipelineState,
    SinkConnector,
    SourceConnector,
)


class CreatePipeline(BaseModel):
    name: str
    space_id: str
    metadata: dict | None = None
    id: str
    created_at: AwareDatetime
    state: PipelineState
    access_token: str


class UpdatePipelineRequest(BaseModel):
    name: str | None = None
    state: str | None = None
    metadata: dict | None = None
    source_connector: SourceConnector | None = None
    sink_connector: SinkConnector | None = None



from .clienterror import ClientError, requests_http


class SpaceNotFoundError(ClientError):
    """Error caused by a space ID not found."""

    def __init__(self, space_id: str, raw_response: requests_http.Response):
        super().__init__(
            detail=f"Space ID {space_id} does not exist",
            status_code=404,
            body=raw_response.text,
            raw_response=raw_response,
        )


class SpaceIsNotEmptyError(ClientError):
    """Error caused by trying to delete a space that is not empty."""

    def __init__(self, raw_response: requests_http.Response):
        super().__init__(
            detail=raw_response.json()["msg"],
            status_code=409,
            body=raw_response.text,
            raw_response=raw_response,
        )


class SpaceUnauthorizedError(ClientError):
    """Space operation not authorized, invalid Personal Access Token"""

    def __init__(self, space_id: str, raw_response: requests_http.Response):
        super().__init__(
            detail=f"Unauthorized request on Space {space_id}, "
            f"Personal Access Token used is invalid",
            status_code=raw_response.status_code,
            body=raw_response.text,
            raw_response=raw_response,
        )



"""GlassFlow Python Client to interact with GlassFlow API"""

from .api_client import APIClient
from .models import errors, responses
from .pipeline import Pipeline
from .secret import Secret
from .space import Space


class GlassFlowClient(APIClient):
    """
    GlassFlow Client to interact with GlassFlow API and manage pipelines
    and other resources

    Attributes:
        client (requests.Session): Session object to make HTTP requests to GlassFlow API
        glassflow_config (GlassFlowConfig): GlassFlow config object to store
            configuration
        organization_id (str): Organization ID of the user. If not provided,
            the default organization will be used

    """

    def __init__(
        self, personal_access_token: str = None, organization_id: str = None
    ) -> None:
        """Create a new GlassFlowClient object

        Args:
            personal_access_token: GlassFlow Personal Access Token
            organization_id: Organization ID of the user. If not provided,
                the default organization will be used
        """
        super().__init__()
        self.personal_access_token = personal_access_token
        self.organization_id = organization_id
        self.request_headers = {"Personal-Access-Token": self.personal_access_token}
        self.request_query_params = {"organization_id": self.organization_id}

    def get_pipeline(self, pipeline_id: str) -> Pipeline:
        """Gets a Pipeline object from the GlassFlow API

        Args:
            pipeline_id: UUID of the pipeline

        Returns:
            Pipeline: Pipeline object from the GlassFlow API

        Raises:
            errors.PipelineNotFoundError: Pipeline does not exist
            errors.PipelineUnauthorizedError: User does not have permission to
                perform the requested operation
            errors.ClientError: GlassFlow Client Error
        """
        return Pipeline(
            personal_access_token=self.personal_access_token,
            id=pipeline_id,
            organization_id=self.organization_id,
        ).fetch()

    def create_pipeline(
        self,
        name: str,
        space_id: str,
        transformation_file: str = None,
        requirements: str = None,
        source_kind: str = None,
        source_config: dict = None,
        sink_kind: str = None,
        sink_config: dict = None,
        env_vars: list[dict[str, str]] = None,
        state: str = "running",
        metadata: dict = None,
    ) -> Pipeline:
        """Creates a new GlassFlow pipeline

        Args:
            name: Name of the pipeline
            space_id: ID of the GlassFlow Space you want to create the pipeline in
            transformation_file: Path to file with transformation function of
                the pipeline.
            requirements: Requirements.txt of the pipeline
            source_kind: Kind of source for the pipeline. If no source is
                provided, the default source will be SDK
            source_config: Configuration of the pipeline's source
            sink_kind: Kind of sink for the pipeline. If no sink is provided,
                the default sink will be SDK
            sink_config: Configuration of the pipeline's sink
            env_vars: Environment variables to pass to the pipeline
            state: State of the pipeline after creation.
                It can be either "running" or "paused"
            metadata: Metadata of the pipeline

        Returns:
            Pipeline: New pipeline

        Raises:
            errors.PipelineUnauthorizedError: User does not have permission to perform
                the requested operation
        """
        return Pipeline(
            name=name,
            space_id=space_id,
            transformation_file=transformation_file,
            requirements=requirements,
            source_kind=source_kind,
            source_config=source_config,
            sink_kind=sink_kind,
            sink_config=sink_config,
            env_vars=env_vars,
            state=state,
            metadata=metadata,
            organization_id=self.organization_id,
            personal_access_token=self.personal_access_token,
        ).create()

    def list_pipelines(
        self, space_ids: list[str] = None
    ) -> responses.ListPipelinesResponse:
        """
        Lists all pipelines in the GlassFlow API

        Args:
            space_ids: List of Space IDs of the pipelines to list.
                If not specified, all the pipelines will be listed.

        Returns:
            responses.ListPipelinesResponse: Response object with the pipelines listed

        Raises:
            errors.PipelineUnauthorizedError: User does not have permission to
                perform the requested operation
        """

        endpoint = "/pipelines"
        query_params = {"space_id": space_ids} if space_ids else {}
        http_res = self._request(
            method="GET", endpoint=endpoint, request_query_params=query_params
        )
        return responses.ListPipelinesResponse(**http_res.json())

    def create_space(
        self,
        name: str,
    ) -> Space:
        """Creates a new Space

        Args:
            name: Name of the Space

        Returns:
            Space: New space

        Raises:
            errors.SpaceUnauthorizedError: User does not have permission to perform
                the requested operation
        """
        return Space(
            name=name,
            personal_access_token=self.personal_access_token,
            organization_id=self.organization_id,
        ).create()

    def list_spaces(self) -> responses.ListSpacesResponse:
        """
        Lists all GlassFlow spaces in the GlassFlow API

        Returns:
            response.ListSpacesResponse: Response object with the spaces listed

        Raises:
            errors.SpaceUnauthorizedError: User does not have permission to perform the
                requested operation
        """

        endpoint = "/spaces"
        http_res = self._request(method="GET", endpoint=endpoint)
        return responses.ListSpacesResponse(**http_res.json())

    def create_secret(self, key: str, value: str) -> Secret:
        """
        Creates a new secret

        Args:
            key: Secret key (must be unique in your organization)
            value: Secret value

        Returns:
            Secret: New secret

        Raises:
            errors.SecretUnauthorizedError: User does not have permission to perform the
                requested operation
        """
        return Secret(
            key=key,
            value=value,
            personal_access_token=self.personal_access_token,
            organization_id=self.organization_id,
        ).create()

    def list_secrets(self) -> responses.ListSecretsResponse:
        """
        Lists all GlassFlow secrets in the GlassFlow API

        Returns:
            responses.ListSecretsResponse: Response object with the secrets listed

        Raises:
            errors.SecretUnauthorizedError: User does not have permission to perform the
                requested operation
        """
        endpoint = "/secrets"
        http_res = self._request(method="GET", endpoint=endpoint)
        return responses.ListSecretsResponse(**http_res.json())

    def _request(
        self,
        method,
        endpoint,
        request_headers=None,
        json=None,
        request_query_params=None,
        files=None,
        data=None,
    ):
        headers = {**self.request_headers, **(request_headers or {})}
        query_params = {**self.request_query_params, **(request_query_params or {})}

        try:
            http_res = super()._request(
                method=method,
                endpoint=endpoint,
                request_headers=headers,
                json=json,
                request_query_params=query_params,
                files=files,
                data=data,
            )
            return http_res
        except errors.UnknownError as e:
            if e.status_code == 401:
                raise errors.UnauthorizedError(e.raw_response) from e
            raise e



from .pipeline import (
    AccessToken,
    ConsumeEventResponse,
    ConsumeFailedResponse,
    ConsumeOutputEvent,
    FunctionLogEntry,
    FunctionLogsResponse,
    ListAccessTokensResponse,
    ListPipelinesResponse,
    PublishEventResponse,
    TestFunctionResponse,
)
from .secret import ListSecretsResponse, Secret
from .space import ListSpacesResponse, Space

__all__ = [
    "ListSpacesResponse",
    "Space",
    "FunctionLogsResponse",
    "FunctionLogEntry",
    "TestFunctionResponse",
    "ListPipelinesResponse",
    "ConsumeEventResponse",
    "ConsumeOutputEvent",
    "PublishEventResponse",
    "ConsumeFailedResponse",
    "ListAccessTokensResponse",
    "AccessToken",
    "Secret",
    "ListSecretsResponse",
]
